# $Id: mail.rb,v 1.6 2004/07/27 04:22:13 tommy Exp $

require "nkf"

def get_from(header)
  h = header.gsub(/\n\s+/, " ").split(/\n/)
  h.each do |hl|
    next unless hl =~ /^From:/i
    f = hl.split(/:/, 2)[1].strip
    if f =~ /<(.*)>/ then
      from = $1.strip
      return from
    end
    if f =~ /\(.*\)/ then
      f = f.gsub(/\(.*\)/, "").strip
    end
    return f
  end
  nil
end

def b64_hdecode(str)
  str.unpack('m')[0]
end

def b64_hencode(str)
  [str].pack("m").gsub(/\n/,"")
end

def qp_hdecode(str)
  str.gsub('_', ' ').gsub(/=([0-9A-F][0-9A-F])/o) do $1.hex.chr end
end

def qp_hencode(str)
  str.gsub(/[=_?]|[^ !-~]/){"=%X"%($&[0])}.gsub(" ","_")
end

def mdecode_token(s)
  nkf_flag = {'iso-2022-jp' => 'J', 'euc-jp' => 'E', 'shift_jis' => 'S', 'us-ascii' => ''}
  if s !~ /\A=\?([a-z0-9_-]+)\?(Q|B)\?([^?]+)\?=\Z/io then
    s
  else
    charset, encoding, text = $1, $2, $3
    c = nkf_flag[charset.downcase]
    if c == nil then return s end
    if encoding.downcase == 'q' then
      s = qp_hdecode(text)
    else
      s = b64_hdecode(text)
    end
    NKF::nkf("-e#{c}m0", s)
  end
end

def get_subject(header)
  h = header.gsub(/\n\s+/, " ").split(/\n/)
  h.each do |hl|
    next unless hl =~ /^Subject:/i
    s = hl.split(/:/, 2)[1].strip
    return s
  end
  nil
end

def trim_subject_prefix(subject, mlname)
  if subject =~ /^(Re:\s*)\[#{mlname}(:\d+)?\]\s*(Re:\s*)?/i then
    subject = $1+$'
  elsif mime_header_decode(subject) =~ /^(Re:\s*)\[#{mlname}(:\d+)?\]\s*(Re:\s*)?/i then
    re = $1
    prefix = $&
    d = split_mime_header_token(subject)
    1.upto(d.size) do |i|
      dd = mime_header_decode(d[0,i].map{|a|a[1]}.join(" "))
      if dd.include? prefix then
        cs, en = d[i-1][2..3]
        d[0,i] = [[dd[prefix.size .. -1], nil, cs, en]]
        break
      end
    end
    d.map! do |t,s,cs,en|
      s ? s : en.downcase=="q" ? "=?#{cs}?#{en}?#{qp_hencode(t)}?=" : "=?#{cs}?#{en}?#{b64_hencode(t)}?="
    end
    subject = re + d.join("\n ")
  end
  subject
end

def split_mime_header_token(str)
  ret = []
  str.split(/\s+/).each do |s|
    if s =~ /^=\?([a-z0-9_-]+)\?(Q|B)\?([^?]+)\?=/i then
      charset, encoding, text = $1, $2, $3
      if encoding.downcase == "q" then
        t = qp_hdecode(text)
      else
        t = b64_hdecode(text)
      end
      ret << [t, s, charset, encoding]
    else
      ret << [s, s, nil, nil]
    end
  end
  ret
end

def mime_header_decode(str)
  return str.gsub(/\s+/, " ").gsub(/\?=\s+=\?/, "?==?").gsub(/=\?[a-z0-9_-]+\?(Q|B)\?[^?]+\?=/i){mdecode_token $&}
end

def mime_header_encode(str)
  if str.length <= 64 and str !~ /[\x80-\xff]/ then
    return str
  end
  w = [""]
  len = 0
  kflag = false
  str.split(//e).each do |c|
    w[-1] << c
    if c.length == 1 then	# ASCII
      len += 3 if kflag		# KANJI-OUT
      len += 1
      kflag = false
    else			# KANJI
      len += 3 unless kflag	# KANJI-IN
      len += c.length
      kflag = true
    end
    if len > 30 then
      len += 3 if kflag		# KANJI-OUT
      len = 0
      w << ""
    end
  end
  w.delete ""

  w2 = []
  w.each do |t|
    w2 << "=?iso-2022-jp?b?"+[NKF::nkf("-Ej", t)].pack("m").chomp+"?="
  end
  w2.join("\n ")
end

if __FILE__ == $0 then
  # get_from test
  raise "error" unless get_from("From: tommy@tmtm.org\n") == "tommy@tmtm.org"
  raise "error" unless get_from("From: TOMITA <tommy@tmtm.org>\n") == "tommy@tmtm.org"
  raise "error" unless get_from("From: tommy@tmtm.org (TOMITA)\n") == "tommy@tmtm.org"

  # b64_hdecode test
  raise "error" unless b64_hdecode("aG9nZWhvZ2U=") == "hogehoge"

  # b64_hencode test
  raise "error" unless b64_hencode("hogehoge") == "aG9nZWhvZ2U="

  # qp_hdecode test
  raise "error" unless qp_hdecode("a_b=3Dc") == "a b=c"

  # qp_hencode test
  raise "error" unless qp_hencode("a?b=c ") == "a=3Fb=3Dc_"

  # mdecode_token test
  raise "error" unless mdecode_token("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=") == "ƥ"

  # get_subject test
  raise "error" unless get_subject("Subject: hoge\n fuga\n") == "hoge fuga"

  # trim_subject_prefix test
  raise "error" unless trim_subject_prefix("test", "mlname") == "test"
  raise "error" unless trim_subject_prefix("[mlname] test", "mlname") == "[mlname] test"
  raise "error" unless trim_subject_prefix("Re: [mlname] test", "mlname") == "Re: test"
  raise "error" unless trim_subject_prefix("Re: [mlname:999] test", "mlname") == "Re: test"
  raise "error" unless trim_subject_prefix("Re: [mlname] Re: test", "mlname") == "Re: test"
  raise "error" unless trim_subject_prefix("=?us-ascii?q?test?=", "mlname") == "=?us-ascii?q?test?="
  raise "error" unless trim_subject_prefix("=?us-ascii?q?[mlname]_test?=", "mlname") == "=?us-ascii?q?[mlname]_test?="
  raise "error" unless trim_subject_prefix("=?us-ascii?q?Re:_[mlname]_test?=", "mlname") == "Re: =?us-ascii?q?test?="
  raise "error" unless trim_subject_prefix("=?us-ascii?q?Re:_[mlname:999]_test?=", "mlname") == "Re: =?us-ascii?q?test?="
  raise "error" unless trim_subject_prefix("=?us-ascii?q?Re:_[mlname]_Re:_test?=", "mlname") == "Re: =?us-ascii?q?test?="
  raise "error" unless trim_subject_prefix("=?us-ascii?q?Re:_[mlna?= =?us-ascii?q?me]_Re:_test?=", "mlname") == "Re: =?us-ascii?q?test?="
  raise "error" unless trim_subject_prefix("=?iso-2022-jp?b?UmU6IFttbG5h?= =?iso-2022-jp?b?bWVdIFJlOiB0ZXN0?=", "mlname") == "Re: =?iso-2022-jp?b?dGVzdA==?="

  # split_mime_header_token test
  raise "error" unless split_mime_header_token("=?us-ascii?q?test?= =?iso-2022-jp?q?hoge?=") == [["test", "=?us-ascii?q?test?=", "us-ascii", "q"], ["hoge", "=?iso-2022-jp?q?hoge?=", "iso-2022-jp", "q"]]

  # mime_header_decode test
  raise "error" unless mime_header_decode("=?us-ascii?q?test?= =?iso-2022-jp?q?hoge?=") == "testhoge"

  # mime_header_encode test
  raise "error" unless mime_header_encode("ƥȥƥ") == "=?iso-2022-jp?b?GyRCJUYlOSVIJUYlOSVIGyhC?="
  raise "error" unless mime_header_encode("ƥabcƥabcƥabcƥabc") == "=?iso-2022-jp?b?GyRCJUYlOSVIGyhCYWJjGyRCJUYlOSVIGyhCYWJjGyRCJUYbKEI=?=\n =?iso-2022-jp?b?GyRCJTklSBsoQmFiYxskQiVGJTklSBsoQmFiYw==?="
end
