# $Id: tml-file.rb,v 1.6 2004/05/09 04:59:16 tommy Exp $
#
# Copyright (C) 2003-2004 TOMITA Masahiro
# 
# tommy@tmtm.org
# 

require "fileutils"

class TML; end

class TML::F

  @@parameters = {
    :lock_dir		=> "lock",
    :deleted_dir	=> "deleted",
  }

  def self.param(id)
    if @@parameters.key? id then
      eval("defined? $#{id}") ? eval("$#{id}") : @@parameters[id]
    else
      TML.param(id)
    end
  end

  def self.dparam(id)
    d = param id
    d[0,1] == "/" ? d : param(:var_dir)+"/"+d
  end

  class Lock
    @@locks = []
    @@lock_dir = nil

    def self.callback(ident, lock_f, orig_lock_f)
      proc do
        File.rename lock_f, orig_lock_f rescue nil
        TML.log "unlock: #{ident}"
        @@locks.delete ident
      end
    end

    def initialize(ident)
      if @@locks.include? ident then
        raise "dead lock detected: #{ident}"
      end
      FileUtils.mkdir_p TML::F.dparam(:lock_dir)
      st = Time.now
      begin
        orig_lock_f = "#{TML::F.dparam :lock_dir}/#{ident}"
        lock_f = "#{orig_lock_f}.#{Time.now.to_i}"
        while Dir.glob("#{orig_lock_f}\0#{orig_lock_f}.*").empty?
          File.open(orig_lock_f, File::WRONLY|File::CREAT|File::EXCL).close
          if Dir.glob("#{orig_lock_f} #{orig_lock_f}.*").size > 1 then
            File.unlink orig_lock_f
          end
        end
        f = Dir.glob("#{orig_lock_f}.*")[0]
        if f then
          if f.split(/\./)[-1].to_i < Time.now.to_i-TML.param(:max_lock_time) then
            File.rename f, orig_lock_f
            TML.log "#{ident}: max_lock_time(#{TML.param(:max_lock_time)}) exceeded"
          end
        end
        File.rename orig_lock_f, lock_f
        TML.log "lock: #{ident}"
        ObjectSpace::define_finalizer(self, Lock.callback(ident, lock_f, orig_lock_f))
        @ident = ident
        @@locks << ident
        @lock_f, @orig_lock_f = lock_f, orig_lock_f
      rescue Errno::ENOENT, Errno::EEXIST
        if st+TML.param(:lock_timeout) < Time.now then
          raise TML::Error.new(:ETIMEDOUT, "lock timed out")
        end
        sleep 1
        retry
      end
    end

    def unlock()
      File.rename @lock_f, @orig_lock_f
      TML.log "unlock: #{@ident}"
      ObjectSpace.undefine_finalizer(self)
      @@locks.delete @ident
    end
  end

  class Meta
    def log_str() nil end
    def type() :hash end

    def initialize(filename)
      @file = filename
      File.open(@file, "w").close unless File.exist? @file
      read_data
    end

    def have?(k)
      @hash.key? item_class.new(k)
    end

    def list()
      @hash.keys
    end

    def add(k, data={})
      if have?(k) then
        raise TML::Error.new(:EEXIST, "#{k}: already exist")
      end
      File.open(@file, "a") do |f|
        if type == :string then
          d = data
        else
          d = data.map{|i,v| "#{i}=#{v}"}.join(",")
        end
        f.puts "#{k}\t#{d}"
      end
      @hash[k] = data
      TML.log "add #{log_str}: #{k}" if log_str
    end

    def delete(k)
      unless have?(k) then
        raise TML::Error.new(:EEXIST, "#{k}: not found")
      end
      @hash.delete item_class.new(k)
      write_data
      TML.log "delete #{log_str}: #{k}" if log_str
    end

    def set(k, data) 
      unless have?(k) then
        raise TML::Error.new(:EEXIST, "#{k}: not found")
      end
      @hash[k] = data
      write_data
      TML.log "set #{log_str}: #{k}" if log_str
    end

    def get(k)
      unless have?(k) then
        raise TML::Error.new(:EEXIST, "#{k}: not found")
      end
      @hash[k]
    end

    def delete_all()
    end

    private

    def item_class() String end

    def read_data()
      @hash = {}
      File.open(@file) do |f|
        f.map do |l|
          next if l =~ /^#/ or l =~ /^\s*$/
          k, data = l.chomp.split(/\s+/,2)
          if type == :string then
            h = data
          else
            h = {}
            if data then
              data.split(/,/).each do |d|
                i, j = d.split(/=/, 2)
                h[i] = j
              end
            end
          end
          @hash[item_class.new(k)] = h
        end
      end
    end

    def write_data()
      bak = "#{@file}.bak"
      new = "#{@file}.new"
      File.open(new, "w") do |f|
        @hash.each do |k,data|
          if type == :string then
            d = data
          else
            d = data.map{|i,v| "#{i}=#{v}"}.join(",")
          end
          f.puts "#{k}\t#{d}"
        end
      end
      FileUtils.cp @file, bak
      FileUtils.cp new, @file
      File.unlink new
    end
  end

  class Addr
    def initialize(str)
      if str.is_a? TML::F::Addr
        @orig_str = str.orig_str
        @cmp_str = str.cmp_str
      else
        @orig_str = str
        @cmp_str = str.sub(/@.*/){$&.downcase}
      end
    end
    attr_reader :orig_str, :cmp_str

    def to_s() @orig_str end
    def to_str() @orig_str end

    def hash()
      self.cmp_str.hash
    end

    def ==(str)
      if str.is_a? TML::F::Addr
        return self.cmp_str == str.cmp_str
      else
        return self.cmp_str == str.sub(/@.*/){$&.downcase}
      end
    end

    alias :=== :==
    alias :eql? :==
  end

  class ML
    def have?(ml)
      File.exist? "#{TML.dparam :ml_dir}/#{ml}"
    end

    def list()
      ret = []
      Dir.glob("#{TML.dparam :ml_dir}/*") do |f|
        ret << f.split(/\/+/)[-1]
      end
      ret
    end

    def add(ml, attr={})
      if have? ml then
        raise TML::Error.new(:EEXIST, "#{ml}: already exist")
      end
      FileUtils.mkdir_p "#{TML.dparam :ml_dir}/#{ml}"
      File.open("#{TML.dparam :ml_dir}/#{ml}/conf", "w") do |f|
        attr.each do |k,v|
          f.puts "#{k}\t#{v}"
        end
      end
      TML.log "add ml: #{ml}"
    end

    def delete(ml)
      unless have? ml then
        raise TML::Error.new(:ENOENT, "#{ml}: not found")
      end
      ts = Time.now.strftime("%Y%m%d%H%M%S")
      FileUtils.mkdir_p TML::F.dparam(:deleted_dir)
      File.rename("#{TML.dparam :ml_dir}/#{ml}", "#{TML::F.dparam :deleted_dir}/#{ml}-#{ts}")
      TML.log "delete ml: #{ml}"
    end

    def set(ml, attr)
      now = get(ml)
      attr.each do |k,v|
        now[k] = v
      end
      file = "#{TML.dparam :ml_dir}/#{ml}/conf"
      bak = "#{file}.bak"
      new = "#{file}.new"
      File.open(new, "w") do |f|
        now.each do |k,v|
          f.puts "#{k}\t#{v}"
        end
      end
      FileUtils.cp file, bak
      FileUtils.cp new, file
      File.unlink new
      TML.log "set ml: #{ml}"
    end

    def get(ml)
      ret = {}
      IO.foreach("#{TML.dparam :ml_dir}/#{ml}/conf") do |l|
        k, v = l.chomp.split(/\s+/, 2)
        ret[k] = (v || "")
      end
      ret
    end
  end

  class Counter
    def have?(ml)
      File.exist? "#{TML.dparam :ml_dir}/#{ml}/counter"
    end

    def list()
      raise "unimplemented"
    end

    def add(ml)
      if have? ml then
        raise TML::Error.new(:EEXIST, "#{ml}: already exist")
      end
      FileUtils.mkdir_p "#{TML.dparam :ml_dir}/#{ml}"
      File.open("#{TML.dparam :ml_dir}/#{ml}/counter", "w") do |f|
        f.puts "0"
      end
      TML.log "add counter: #{ml}"
    end

    def delete(ml)
    end

    def set(ml, cnt)
      File.open("#{TML.dparam :ml_dir}/#{ml}/counter", "w") do |f|
        f.puts cnt
      end
    end

    def get(ml)
      File.open("#{TML.dparam :ml_dir}/#{ml}/counter") do |f|
        f.gets.to_i
      end
    end
  end

  class Passwd < Meta
    def log_str() "passwd" end

    def initialize()
      FileUtils.mkdir_p TML.dparam(:etc_dir)
      super "#{TML.dparam :etc_dir}/passwd"
    end

    def add(addr, pw)
      super(addr, {"passwd"=>pw})
    end

    def set(addr, pw)
      super(addr, {"passwd"=>pw})
    end

    def auth(addr, pw)
      begin
        p = get(addr)["passwd"]
        if p == pw then
          return true
        end
      rescue TML::Error => err
        raise unless err.id == :ENOENT
      end
      TML.log "auth failed: #{addr}"
      return false
    end
  end

  class Member < Meta
    def log_str() "member" end
    def item_class() Addr end
    def initialize(ml)
      super "#{TML.dparam :ml_dir}/#{ml}/member"
    end
    def add(addr)
      super Addr.new(addr)
    end
    def list()
      super.map{|a|a.to_s}
    end
    undef set
  end

  class Admin < Meta
    def log_str() "admin" end
    def item_class() Addr end
    def initialize(ml)
      super "#{TML.dparam :ml_dir}/#{ml}/admin"
    end
    def add(addr)
      super Addr.new(addr)
    end
    def list()
      super.map{|a|a.to_s}
    end
    undef set
  end

  class Child < Meta
    def log_str() "child" end
    def initialize(ml)
      @ml = ml
      super "#{TML.dparam :ml_dir}/#{ml}/child"
    end
    def add(ml)
      super ml
      Parent.new(ml).add(@ml)
    end
    def delete(ml)
      super ml
      Parent.new(ml).delete(@ml)
    end
  end

  class Parent < Meta
    def log_str() "parent" end
    def initialize(ml)
      super "#{TML.dparam :ml_dir}/#{ml}/parent"
    end
  end

  class Spool < Meta
    def log_str() "spool" end
    def type() :string end

    def initialize(ml)
      @ml = ml
      super "#{TML.dparam :ml_dir}/#{ml}/summary"
    end

    def spool(seq, subject, body)
      add(seq.to_s, subject)
      FileUtils.mkdir_p "#{TML.dparam :ml_dir}/#{@ml}/spool"
      File.open("#{TML.dparam :ml_dir}/#{@ml}/spool/#{seq}", "w") do |f|
        f.puts body
      end
    end

    def copy_messages(msgs, dest)
      msgs.each do |f|
        fn = "#{TML.dparam :ml_dir}/#{@ml}/spool/#{f}"
        begin
          FileUtils.cp(fn, dest)
        rescue Errno::ENOENT
        end
      end
    end

    def summary_list(msgs=nil)
      if msgs == nil then
        msgs = list.sort{|a,b|a.to_i<=>b.to_i}
      end
      msgs.map{|i| "#{i} #{get(i.to_s)}"}.join("\n")
    end
  end

  class Work
    def initialize(ml)
      @dir = "#{TML.dparam :ml_dir}/#{ml}/tmp"
    end

    def add(key, data)
      FileUtils.mkdir_p @dir
      File.open("#{@dir}/#{key}", "w") do |f|
        f.puts data
      end
    end

    def get(key)
      unless File.exist? "#{@dir}/#{key}" then
        raise TML::Error.new(:ENOENT, "#{key}: not found")
      end
      File.open("#{@dir}/#{key}") do |f|
        f.read
      end
    end

    def delete(key)
      File.unlink "#{@dir}/#{key}" rescue nil
    end

    def expire()
      now = Time.now
      Dir.glob("#{@dir}/*").each do |f|
        if File.mtime(f)+24*60*60 < now then
          File.unlink f
        end
      end
    end
  end
end

# Test
#
if __FILE__ == $0 then

  require "tml.rb"
  def TML.log(a) puts a end
  $var_dir = "/tmp/tml-test"

  # Lock test
  l = TML::F::Lock.new("hoge")
  begin
    TML::F::Lock.new("hoge")
    raise "error!"
  rescue => e
    raise unless e.to_s == "dead lock detected: hoge"
  end
  l.unlock
  TML::F::Lock.new("hoge").unlock

  # ML test
  ml = TML::F::ML.new
  ml.add("hoge")
  raise "error!" unless ml.list == ["hoge"]
  ml.set("hoge", {"overview"=>"HOGEHOGE", "not_member"=>"true"})
  h = ml.get("hoge")
  raise "error!" if h["overview"] != "HOGEHOGE" or h["not_member"] != "true"

  # Passwd test
  pw = TML::F::Passwd.new
  pw.add("hoge", "hogepw")
  raise "error!" unless pw.list == ["hoge"]
  pw.auth("hoge", "hogepw")
  pw.set("hoge", "hogepw2")
  pw.auth("hoge", "hogepw2")
  pw.delete("hoge")
  raise "error!" if pw.have? "hoge"

  # Member test
  mm = TML::F::Member.new("hoge")
  mm.add("hogehoge")
  raise "error!" unless mm.list == ["hogehoge"]
  mm.delete("hogehoge")
  raise "error!" if mm.have? "hogehoge"

  # Admin test
  ad = TML::F::Admin.new("hoge")
  ad.add("xxxx")
  raise "error!" unless ad.list == ["xxxx"]
  ad.delete("xxxx")
  raise "error!" if ad.have? "xxxx"

  # Spool test
  sp = TML::F::Spool.new("hoge")
  sp.spool(1, "subject1", "body1\nbody1\n")
  sp.spool(2, "subject2", "body2\nbody2\n")
  raise "error!" if sp.summary_list([1, 2]) != "1 subject1\n2 subject2"

  # Tree test
  ml.add("fuga")
  ch = TML::F::Child.new("hoge")
  ch.add("fuga")
  raise "error!" unless ch.list == ["fuga"]
  raise "error!" unless TML::F::Parent.new("fuga").list == ["hoge"]
  ch.delete("fuga")
  raise "error!" if ch.have? "fuga"
  ml.delete("fuga")
  ml.delete("hoge")
  raise "error!" if ml.have? "hoge"
end
