# $Id: tml.rb,v 1.40 2004/07/04 16:57:31 tommy Exp $
#
# Copyright (C) 2003-2004 TOMITA Masahiro
# tommy@tmtm.org
#

TML_VERSION = "0.5"

class TML

  @@msg_cnt = 0
  @@log_f = nil
  @@cache = {}
  @@parameters = {
    :max_lock_time	=> 30*60,
    :lock_timeout	=> 5*60,
    :tml_dir		=> "/usr/local/tml",
    :var_dir		=> "/var/spool/tml",
    :ml_dir		=> "ml",
    :etc_dir		=> "etc",
    :log_file		=> "log",
    :smtp_server	=> "localhost",
    :smtp_port		=> 25,
    :auto_aliases	=> true,
    :newaliases		=> nil,
    :strip_received	=> false,
    :logging		=> true,
    :archive_command	=> "tar cf - %s | bzip2",
    :archive_ext	=> "tar.bz2",
    :passwd_char	=> "abcdefghjkmnprstuvwxyzABCDEFGHJKLMNOPRSTUVWXYZ2345678",
    :passwd_length	=> 8,
  }

  @@attributes = {
    :sender		=> [:String, nil],
    :overview		=> [:String, nil],
    :not_member		=> [:Boolean, false],
    :subject_mlname	=> [:Boolean, true],
    :counter_width	=> [:Integer, 5],
    :remote_command	=> [:Boolean, false],
    :remote_admin	=> [:Boolean, false],
    :spool		=> [:Boolean, true],
    :web_interface	=> [:String, nil],
    :need_confirm_command => [:String, "subscribe,admin"],
    :allowed_command	=> [:String, "all"],
    :anonymous_command	=> [:String, "subscribe,confirm"],
    :admin_command	=> [:String, "all"],
    :logging		=> [:Boolean, true],
    :replyto_type	=> [:String, "ml"],
  }

  @@access_priv = [
    :permit, :permit_post, :permit_subscribe,
    :reject, :reject_post, :reject_subscribe,
  ]

  class Error < StandardError
    def initialize(id, msg)
      TML.log msg
      @id, @msg = id, msg
      super msg
    end
    attr_reader :id, :msg
  end

  class Attr
    def initialize(name, type)
      @name, @type = name, type
      @value = nil
    end

    attr_reader :name, :type, :value

    def set(value)
      case @type
      when :Integer
	if value.kind_of? Integer then
	  @value = value
	elsif value.kind_of? String then
	  unless value =~ /^\d+$/ then
	    raise Error.new(:EINVAL, "#{@name}=#{value}: integer value required")
	  end
	  @value = value.to_i
	else
	  raise Error.new(:EINVAL, "#{@name}=#{value}: invalid type")
	end
      when :Boolean
	if value.kind_of? TrueClass or value.kind_of? FalseClass then
	  @value = value
	elsif value.kind_of? NilClass then
	  @value = false
	elsif value.kind_of? String then
	  case value.downcase
	  when "true", "yes", "y"
	    @value = true
	  when "false", "no", "n"
	    @value = false
	  else
	    raise Error.new(:EINVAL, "#{@name}=#{value}: invalid value")
	  end
	else
	  raise Error.new(:EINVAL, "#{@name}=#{value}: invalid type")
	end
      when :String
	@value = value
      else
	raise "#{@name}=#{value}: invalid type"
      end
    end
  end

  def self.db_class()
    if defined? $database and $database == "mysql" then
      require "#{self.dparam(:tml_dir)}/tml-mysql.rb"
      TML::Mysql
    else
      require "#{self.dparam(:tml_dir)}/tml-file.rb"
      TML::F
    end
  end

  def self.spool_class()
    if defined? $spool_db then
      if $spool_db == "mysql" then
        require "tml-mysql.rb"
        TML::Mysql
      else
        require "tml-file.rb"
        TML::F
      end
    else
      db_class
    end
  end

  def self.param(id)
    eval("defined? $#{id}") ? eval("$#{id}") : @@parameters[id]
  end

  def self.dparam(id)
    d = param id
    d[0,1] == "/" ? d : param(:var_dir)+"/"+d
  end

  def self.log(msg)
    return unless param(:logging)
    unless @@log_f then
      @@log_f = File.open(dparam(:log_file), "a")
      @@log_f.sync = true
    end
    @@log_f.puts "#{TML.log_date} #{File.basename $0}[#{$$}] #{msg}"
  end

  def self.log_date()
    Time.now.strftime("%Y-%m-%d %H:%M:%S")
  end

  def self.list()
    TML.db_class::ML.new.list
  end

  def self.create(mlname, admin)
    TML.log "create #{mlname}"
    TML.db_class::ML.new.add mlname
    TML.db_class::Counter.new.add mlname
    tml = TML[mlname]
    @@attributes.each do |k,v|
      val = eval("defined? $default_#{k}") ? eval("$default_#{k}") : v[1]
      if val.is_a? Proc then val = val.call end
      tml.attr[k].set(val)
    end
    tml.attr[:sender].set "#{mlname}-admin@#{$domain}"
    tml.write_conf
    tml.add_admin admin
    TML.add_alias mlname if TML.param(:auto_aliases)
    tml
  end

  def self.drop(mlname)
    lock = TML.db_class::Lock.new "ml:#{mlname}"
    TML.log "drop #{mlname}"
    TML.del_alias mlname if TML.param(:auto_aliases)
    TML.db_class::Member.new(mlname).delete_all
    TML.db_class::Admin.new(mlname).delete_all
    TML.db_class::Child.new(mlname).delete_all
    TML.db_class::Parent.new(mlname).delete_all
    TML.db_class::Counter.new.delete mlname rescue nil
    TML.db_class::Spool.new(mlname).delete_all
    TML.db_class::ML.new.delete mlname
    lock.unlock
  end

  def self.exist?(mlname)
    TML.db_class::ML.new.have? mlname
  end

  def self.add_alias(mlname)
    require "fileutils"
    TML.log "add alias"
    lock = TML.db_class::Lock.new "etc:alias"
    FileUtils.mkdir_p TML.dparam(:etc_dir)
    file = "#{TML.dparam :etc_dir}/aliases"
    File.open(file, "a") do |f|
      f.puts <<EOS
#{mlname}: "|#{TML.dparam :tml_dir}/tml #{mlname}"
#{mlname}-admin: :include:#{TML.dparam :ml_dir}/#{mlname}/admin
#{mlname}-ctl: "|#{TML.dparam :tml_dir}/tmlctl #{mlname}"
EOS
    end
    system TML.param(:newaliases) if TML.param(:newaliases)
    lock.unlock
  end

  def self.del_alias(mlname)
    TML.log "delete alias"
    lock = TML.db_class::Lock.new "etc:alias"
    file = "#{TML.dparam :etc_dir}/aliases"
    File.open(file+".new", "w") do |f|
      IO.foreach(file) do |line|
	next if line =~ /^#{Regexp.escape mlname}(-ctl|-admin)?:/
	f.puts line
      end
    end
    File.rename(file+".new", file)
    system TML.param(:newaliases) if TML.param(:newaliases)
    lock.unlock
  end

  def self.new(mlname)
    if @@cache.has_key? mlname then
      return @@cache[mlname]
    end
    tml = super mlname
    @@cache[mlname] = tml
    tml
  end

  def self.[](mlname)
    TML.new mlname
  end

  def initialize(mlname)
    unless TML.exist? mlname then
      err :ENOENT, "#{mlname}: not exist"
    end
    @lock_f = TML.db_class::Lock.new "ml:#{mlname}"
    @name = mlname
    @attr = {}
    @@attributes.each do |n,t|
      @attr[n] = Attr.new(n, t[0])
    end
    @ml = TML.db_class::ML.new
    @counter = TML.db_class::Counter.new
    @member = TML.db_class::Member.new mlname
    @admin = TML.db_class::Admin.new mlname
    @child = TML.db_class::Child.new mlname
    @parent = TML.db_class::Parent.new mlname
    @work = TML.db_class::Work.new mlname
    @spool = TML.spool_class::Spool.new mlname

    read_conf

    @work.expire
  end

  def close()
    @lock_f.unlock
    @name = nil
  end

  attr_reader :name, :attr

  def log(msg)
    require "fileutils"
    TML.log "#{@name}: #{msg}"
    return unless get_attr(:logging)
    FileUtils.mkdir_p "#{TML.dparam :ml_dir}/#{@name}"
    File.open("#{TML.dparam :ml_dir}/#{@name}/log", "a") do |f|
      f.puts "#{TML.log_date} #{File.basename $0}[#{$$}] #{msg}"
    end
  end

  # Counter

  def counter()
    @counter.get(@name)
  end

  def counter=(cnt)
    log "set counter: #{cnt}"
    @counter.set(@name, cnt)
  end

  def countup()
    c = @counter.get(@name)
    c += 1
    @counter.set(@name, c)
    c
  end

  # Member

  def members()
    @member.list
  end

  def all_members()
    m = []
    m += members
    children.each do |c|
      m += TML[c].all_members
    end
    m.uniq
  end

  def have?(addr)
    return true if @member.have? addr
    children.each do |c|
      return true if TML[c].have? addr
    end
    false
  end

  def add(*addrs)
    log "add member: #{addrs.join(",")}"
    addrs.each do |a|
      @member.add a
    end
  end

  def bye(*addrs)
    log "bye member: #{addrs.join(",")}"
    addrs.each do |a|
      @member.delete a
    end
  end

  # ML tree

  def children()
    @child.list
  end

  def addml(ml)
    log "add child: #{ml}"
    ml = ml.name if ml.kind_of? TML
    unless TML.exist? ml then
      err :ENOENT, "#{ml}: not found"
    end
    @child.add ml
  end

  def byeml(ml)
    log "bye child: #{ml}"
    ml = ml.name if ml.kind_of? TML
    @child.delete ml
  end

  def parents()
    @parent.list
  end

  # Administrator

  def admins()
    @admin.list
  end

  def admin?(addr)
    return true if @admin.have? addr
  end

  def add_admin(addr)
    log "add admin: #{addr}"
    @admin.add addr
  end

  def bye_admin(addr)
    log "bye admin: #{addr}"
    if @admin.list.size == 1 then
      err :EBUSY, "one admin is needed at least "
    end
    @admin.delete addr
  end

  # Access

  def access()
    return @access if @access
    @access = []
    if File.exist? "#{TML.dparam :ml_dir}/#{@name}/access" then
      af = "#{TML.dparam :ml_dir}/#{@name}/access"
    elsif File.exist? "#{TML.dparam :etc_dir}/access" then
      af = "#{TML.dparam :etc_dir}/access"
    else
      return @access
    end
    IO.foreach(af) do |l|
      next if l[0] == ?\# or l =~ /^\s*$/
      re, priv = l.chomp.split(/\t+/, 2)
      unless @@access_priv.include? priv.intern then
	raise Error.new(:EINVAL, "#{priv}: unknown privilege")
      end
      @access << [Regexp.new(re, Regexp::IGNORECASE), priv.intern]
    end
    @access
  end

  def check_subscribe(mail)
    access.each do |re, priv|
      if mail =~ re then
	if priv == :permit or priv == :permit_subscribe then
	  return true
	end
	if priv == :reject or priv == :reject_subscribe then
	  log "reject to subscribe: #{mail}"
	  return false
	end
      end
    end
    true
  end

  def check_post(mail)
    access.each do |re, priv|
      if mail =~ re then
	if priv == :permit or priv == :permit_post then
	  return true
	end
	if priv == :reject or priv == :reject_post then
	  log "reject to post: #{mail}"
	  return false
	end
      end
    end
    true
  end

  # Attribute

  def set_attr(attr, value)
    log "set attribute: #{attr}=#{value}"
    attr = attr.intern if attr.kind_of? String
    unless @@attributes.key? attr then
      err :EINVAL, "#{attr}: unknown attribute"
    end
    @attr[attr].set value
    write_conf
  end

  def get_attr(attr=nil)
    if attr then
      attr = attr.intern if attr.kind_of? String
      unless @@attributes.key? attr then
	err :EINVAL, "#{attr}: unknown attribute"
      end
      return @attr[attr].value
    else
      ret = {}
      @attr.each do |n, a|
	ret[n] = a.value
      end
      return ret
    end
  end

  def read_conf()
    @ml.get(@name).each do |k,v|
      id = k.intern
      unless @@attributes.key? id then
	warn "unknown attribute: #{k}"
      else
	@attr[id].set v
      end
    end
  end

  def write_conf()
    log "write config"
    hash = {}
    @attr.each do |k,v|
      hash[k.to_s] = v.value.to_s
    end
    @ml.set(@name, hash)
  end

  # Spooling

  def spool(seq, subject, msg)
    log "spool #{seq} #{subject}"
    @spool.spool(seq, subject, msg)
  end

  def range(s, e=nil)
    list = @spool.list.map{|i| i.to_i}
    e ||= s
    list.delete_if{|i| i < s}
    list.delete_if{|i| i > e}
    list
  end

  def archive(msg)
    require "tmpdir"
    require "fileutils"

    workdir = Dir.tmpdir+"/tml-#{$$}"
    begin
      Dir.mkdir workdir
    rescue Errno::EEXIST
      workdir.succ
      retry
    end
    begin
      work2 = "#{workdir}/messages"
      Dir.mkdir work2
      @spool.copy_messages msg, work2
      arc = "#{workdir}/messages.#{TML.param(:archive_ext)}"
      cmd = sprintf(TML.param(:archive_command), msg.join(" "))
      system "cd #{work2}; #{cmd} > #{arc}"
      if $?.exitstatus != 0 then
        err :EINVAL, "archive error"
      end
      return [File.open(arc){|f|f.read}].pack("m")
    ensure
      FileUtils.rm_rf workdir
    end
  end

  def summary_all()
    @spool.summary_list
  end

  def summary(s, e=nil)
    @spool.summary_list(range(s, e))
  end

  # Password

  def self.make_passwd()
    pc = param(:passwd_char)
    ret = ""
    param(:passwd_length).times do
      ret << pc[rand(pc.length),1]
    end
    ret
  end

  def self.set_passwd(addr, passwd)
    pw = TML.db_class::Passwd.new
    if pw.have? addr then
      pw.set(addr, passwd)
    else
      pw.add(addr, passwd)
    end
  end

  def self.auth(addr, passwd)
    TML.db_class::Passwd.new.auth(addr, passwd)
  end

  # Other

  def self.template(mlname, name)
    if File.exist? "#{TML.dparam :ml_dir}/#{mlname}/templates/#{name}" then
      return File.open("#{TML.dparam :ml_dir}/#{mlname}/templtates/#{name}"){|f| f.read}
    elsif File.exist? "#{TML.dparam :var_dir}/templates/#{name}" then
      return File.open("#{TML.dparam :var_dir}/templates/#{name}"){|f| f.read}
    elsif File.exist? "#{TML.dparam :tml_dir}/templates/#{name}" then
      return File.open("#{TML.dparam :tml_dir}/templates/#{name}"){|f| f.read}
    else
      return nil
    end
  end

  def self.send_msg(mlname, sender, from, to, templ, hash={})
    require "net/smtp"
    require "date"
    hash[:from] = from unless hash.key? :from
    hash[:to] = to.is_a?(Array) ? to[0] : to unless hash.key? :to
    hash[:mlname] = mlname unless hash.key? :mlname
    hash[:adminaddr] = "#{mlname}-admin@#{$domain}" unless hash.key? :adminaddr
    hash[:ctladdr] = "#{mlname}-ctl@#{$domain}" unless hash.key? :ctladdr
    msg = template(mlname, templ)
    msg.gsub!(/%\{(\w+)\}/){hash[$1.intern]}
    date = DateTime.now.strftime("Date: %a, %d %b %Y %H:%M:%S %z")+"\n"
    msgid = "Message-Id: <#{Time.now.to_i}.#{$$}_#{@@msg_cnt}.#{mlname}.tml@#{$domain}>\n"
    msg = date + msgid + msg
    Net::SMTP.start(TML.param(:smtp_server), TML.param(:smtp_port)) do |smtp|
      smtp.send_mail(msg, sender, to)
    end
    @@msg_cnt += 1
  end

  def send_msg(sender, from, to, templ, hash={})
    TML.send_msg self.name, sender, from, to, templ, hash
  end

  def work_set(key, data)
    @work.add key, data
  end

  def work_get(key)
    @work.get key
  end

  def work_delete(key)
    @work.delete key
  end

  private

  def err(errid, msg)
    raise TML::Error.new(errid, msg)
  end

  def warn(msg)
    $stderr.puts "warning: "+msg
  end

end
